##
# Copyright 2023 Physics Front LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

import io, os, setuptools, sys


# Package metadata.

name = "physicsfront-qiskit"
description = "Utility package for qiskit"

# Should be one of:
# 'Development Status :: 3 - Alpha'
# 'Development Status :: 4 - Beta'
# 'Development Status :: 5 - Production/Stable'
release_status = 'Development Status :: 3 - Alpha'
##
# These dependencies and python_requires are rough values based on the
# limited testing in google colab (Jan 2023).
##
dependencies = [
    "matplotlib >= 3.2.2",
    "qiskit >= 0.39.4",
    "pylatexenc >= 2.10",
]
#extras = {}
python_requires = ">=3.8"

# Setup boilerplate below this line.

package_root = os.path.abspath (os.path.dirname (__file__))

version = "0.1.0"

readme_filename = os.path.join (package_root, "README.md")
with io.open (readme_filename, encoding="utf-8") as readme_file:
    readme = readme_file.read ()

# Only include packages under the 'physicsfront' namespace. Do not include tests, benchmarks, etc.
packages = [
    package for package in setuptools.find_packages () if package.startswith("physicsfront")
]

# Determine which namespaces are needed.
namespaces = ["physicsfront"]
#if "physicsfront.cloud" in packages:
#    namespaces.append("physicsfront.cloud")

# for 'ap' dev cycle
if __name__ == '__main__' and len (sys.argv) == 1:
    sys.argv.append ('check')

setuptools.setup (
    name=name,
    version=version,
    description=description,
    long_description=readme,
    author="Physics Front LLC",
    author_email="info@physicsfront.com",
    license="Apache 2.0",
    url="https://github.com/sam-pf/pf-qiskit",
    classifiers=[
        release_status,
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Security :: Cryptography",
    ],
    platforms="Posix; MacOS X; Windows",
    packages=packages,
    namespace_packages=namespaces,
    install_requires=dependencies,
    #extras_require=extras,
    python_requires=python_requires,
    include_package_data=True,
    #zip_safe=False,
)



##
# Copyright 2023 Physics Front LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

#from qiskit.tools.monitor import job_monitor # pylint: disable=W0611
#from qiskit.visualization import plot_histogram # pylint: disable=W0611

def qc_entangled_two_qbits (kind = 1, statevector = False, measure = True):
    """
    :param statevector:  Only for simulator run.
    """
    from qiskit import QuantumCircuit
    if statevector:
        # This is for 'qc.save_statevector'.
        from qiskit.providers.aer import Aer # pylint: disable=E0401,E0611,W0611
    qc = QuantumCircuit (2)
    if kind == 0: # symmetric Bell sate: |00> + |11>
        iv0 = [1, 0]
        iv1 = [1, 0]
    elif kind == 1: # EPR: -|01> + |10>
        iv0 = [0, 1]
        iv1 = [0, 1]
    elif kind == 2: # |00> - |11>
        iv0 = [0, 1]
        iv1 = [1, 0]
    elif kind == 3: # |01> + |10>
        iv0 = [1, 0]
        iv1 = [0, 1]
    else:
        iv0, iv1 = kind
    qc.initialize (iv0, 0) # pylint: disable=E1101
    qc.initialize (iv1, 1) # pylint: disable=E1101
    qc.h (0)
    qc.cx (0, 1)
    if statevector:
        qc.save_statevector () # pylint: disable=E1101
        ##
        # Later, do 'j.result ().get_statevector ()' where j is job (returned
        # by run_quantum_simulator)
        ##
    if measure:
        qc.measure_all ()
    return qc

def qc_for_random_bits (statevector = False, measure = True):
    """
    :param statevector:  Only for simulator run.
    """
    from qiskit import QuantumCircuit
    if statevector:
        # This is for 'qc.save_statevector'.
        from qiskit.providers.aer import Aer # pylint: disable=E0401,E0611,W0611
    qc = QuantumCircuit (1)
    qc.initialize ([1, 0], 0) # pylint: disable=E1101
    qc.h (0)
    if statevector:
        qc.save_statevector () # pylint: disable=E1101
    if measure:
        qc.measure_all ()
    return qc

def run_quantum (qc, hub = 'ibm-q', shots = 2000, memory = True,
                 backend = None):
    """
    Runs a real quantum computer.  This is how the run result can be processed.

    ```
    qc.draw ()
    j = run_quantum (qc)
    job_monitor (j, interval = 2)
    r = j.result ()
    plot_histogram (r.get_counts ())
    m = r.get_memory ()
    ```
    """
    from qiskit import IBMQ, transpile
    from qiskit.providers.ibmq import least_busy
    if not IBMQ.active_account ():
        raise Exception ("No IBMQ acount is active.")
    provider = IBMQ.get_provider (hub = hub)
    n = qc.num_qubits
    if not backend:
        backend = least_busy (provider.backends
                (filters = lambda x: x.configuration().n_qubits >= n and
                                     not x.configuration().simulator and
                                     x.status().operational == True))
        print ("backend: automatically determined as the least busy one:",
               backend)
    transpiled_qc = transpile (qc, backend, optimization_level=3)
    return backend.run (transpiled_qc, shots = shots, memory = memory)

def run_quantum_simulator (qc, shots = 2000, memory = True, seed = 100):
    """
    Runs a quantum simulator.

    See :func:`run_quantum` for how to process the run result.

    In a simulator run, it is also possible to do

        `r.get_statevector ()`

    assuming that the statevector was saved in the qc.

    :param seed:  The seed value is intentionally fixed at an integer value
        to emphasize the deterministic nature of the simulator. If seed is
        set to None, then a seed will be reset for each run by the system.
    """
    from qiskit import Aer, assemble
    o = assemble (qc)
    sim = Aer.get_backend ('aer_simulator')
    sim.set_option ('seed_simulator', seed)
    return sim.run (o, shots = shots, memory = memory)



##
# Copyright 2023 Physics Front LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

def _install_pip_packages (reload = False, quiet = False):
    import subprocess, sys
    if 'qiskit' in sys.modules and not reload:
        raise Exception ("The initializtion function is called with the "
                         "qiskit module already imported, while reload "
                         "is false.")
    pkgs = ["qiskit", "pylatexenc"]
    if not quiet:
        print ("== Installing required pip packages:", pkgs, '...', end = '',
               flush = True)
    r = subprocess.run ([sys.executable, "-m", "pip", "install"] + pkgs, # pylint: disable=W1510
                        capture_output = True)
    rv = r.returncode
    if not rv:
        if not quiet:
            print ("OK!")
        return
    if not quiet:
        print ("\r", end = '')
    print ("** Error while installing required pip packages:", pkgs)
    print (r.stderr.decode ('utf-8'))
    raise Exception ("pip returned error %d" % (rv,))

def _setup_account (reload = False, quiet = False):
    from qiskit import IBMQ
    if not quiet:
        print ("== Account setup (you may be prompted for an API TOKEN) ...")
    d = IBMQ.active_account ()
    old_token = ''
    if d:
        if reload:
            old_token = d ['token'] if d else ''
            IBMQ.disable_account ()
        else:
            raise Exception ("The initializtion fucntion is called with an "
                             "active IBMQ account, while reload is false.")
    import IPython
    from IPython.core.display import display # pylint: disable=E0611
    from google.colab import output # pylint: disable=E0401,E0611
    display (IPython.display.Javascript (f'window._key = "{old_token}"'))
    display (IPython.display.Javascript ('''
        window._key = prompt ("Please enter your IBMQ API TOKEN:", window._key)
    '''))
    token = output.eval_js ('_key')
    # for safety; even if javascript is sandboxed per cell
    output.eval_js ('delete window._key')
    IBMQ.enable_account (token)
    if not IBMQ.active_account ():
        raise Exception ("Failed to set up an IBMQ account.")
    if not quiet:
        print ("== Account setup ... OK!")

def init (reload = False, quiet = False):
    _install_pip_packages (reload = reload, quiet = quiet)
    _setup_account (reload = reload, quiet = quiet)
